// code to solve exercises for IMPRS topology course

#include <iostream>
#include <iomanip>
#include <fstream>
#include <complex>
#include <vector>
#include <assert.h>
#include <math.h>
#include "matrix.h"
#ifndef NO_MPI
    #include <mpi.h>
#endif

//physics parameters
#define NUMMU   200
#define MUMIN   0.
#define MUMAX   4.
#define NORB    50 // standard matrix size, NORB/2 = number of sites in Kitaev chain
#define THOP    1.0
#define DELTA   1.0

using namespace std;

typedef complex<double> cdouble;
typedef matrix<cdouble> cmat;
cdouble II(0,1);



// LAPACK stuff
extern "C" void zheev_(char* jobz, char* uplo, int* N, cdouble* H, int* LDA, double* W, cdouble* work, int* lwork, double* rwork, int *info);

int     matsize = NORB;
int     lwork = 2*(2*NORB-1);
double  rwork[3*NORB-2];
cdouble work[2*(2*NORB-1)];
char    jobz = 'V';
char    uplo = 'U';
int     info;


void diagonalize(cmat &H, vector<double> &evals)
{
	zheev_(&jobz, &uplo, &matsize, H.ptr(), &matsize, &evals[0], &work[0], &lwork, &rwork[0], &info);
	assert(!info);
}

void times(cmat &A, cmat &B, cmat &C)
{
    for(int i=0; i<NORB; ++i)
        for(int j=0; j<NORB; ++j)
        {
            C(i,j) = 0.;
            for(int k=0; k<NORB; ++k)
            {
                C(i,j) += A(i,k)*B(k,j);
            }
        }
}

void times_dn(cmat &A, cmat &B, cmat &C)
{
	for(int i=0; i<NORB; ++i)
        for(int j=0; j<NORB; ++j)
        {
            C(i,j) = 0.;
            for(int k=0; k<NORB; ++k)
            {
                C(i,j) += conj(A(k,i))*B(k,j);
            }
        }
}

void times_nd(cmat &A, cmat &B, cmat &C)
{
	for(int i=0; i<NORB; ++i)
        for(int j=0; j<NORB; ++j)
        {
            C(i,j) = 0.;
            for(int k=0; k<NORB; ++k)
            {
                C(i,j) += A(i,k)*conj(B(j,k));
            }
        }
}

inline double get_ek(double kx, double ky)
{
	return (kx*kx + ky*ky);
}


int main(int argc, char * argv[])
{
    //************** MPI INIT ***************************
  	int numprocs=1, myrank=0, namelen;
    
#ifndef NO_MPI
  	char processor_name[MPI_MAX_PROCESSOR_NAME];
  	MPI_Init(&argc, &argv);
  	MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
  	MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
  	MPI_Get_processor_name(processor_name, &namelen);
    
	//cout << "Process " << myrank << " on " << processor_name << " out of " << numprocs << " says hello." << endl;
	MPI_Barrier(MPI_COMM_WORLD);
    
	MPI_Datatype ttype  = MPI_DOUBLE;
    
#endif
	if(myrank==0) cout << "\n\tProgram running on " << numprocs << " processors." << endl;
    
	//if(myrank==0)
	//	printf("\n\n");
    
#ifndef NO_MPI
	MPI_Barrier(MPI_COMM_WORLD);
#endif

    
    ofstream outfile;
    ofstream outstates;
    ofstream outstates2;
    
    if(myrank==0)
    {
        outfile.open("bands.dat");
        outfile << setprecision(8);
        outstates.open("states.dat");
        outstates << setprecision(8);
        outstates2.open("states2.dat");
        outstates2 << setprecision(8);
    }
 
    
    cmat Hmat(NORB, NORB);
    vector<double> evals(NORB);
 
    int kcounter = 0;
    
    double dmu = (MUMAX-MUMIN)/NUMMU;
    
    for(double mu=MUMIN; mu<=MUMAX; mu+=dmu)
    {
        
        Hmat.zero();

        //diagonal blocks
        for(int i=0; i<NORB/2; ++i)
        {
            Hmat(2*i,2*i) = -mu;
            Hmat(2*i+1,2*i+1) = mu;
        }
        
        // these are the off-diagonal blocks
        for(int i=0; i<NORB/2-1; ++i)
        {
            Hmat(2*i,2*i+2) = -THOP;
            Hmat(2*i,2*i+3) = -DELTA;
            Hmat(2*i+1,2*i+3) = THOP;
            Hmat(2*i+1,2*i+2) = DELTA;
            Hmat(2*i+2,2*i) = -THOP;
            Hmat(2*i+3,2*i) = -DELTA;
            Hmat(2*i+3,2*i+1) = THOP;
            Hmat(2*i+2,2*i+1) = DELTA;
        }
        
        /*
        if(myrank==0 && kcounter==50)
        {
            for(int i=0; i<NORB; ++i)
            {
                for(int j=0; j<NORB; ++j)
                    cout << Hmat(i,j) << '\t';
                
                cout << endl;
            }
        }
         */
        
        diagonalize(Hmat, evals);
        
        if(myrank==0)
        {
            outfile << kcounter << '\t' << mu;
            for(int i=0; i<NORB; ++i)
            {
                outfile << '\t' << evals[i];
            }
            outfile << endl;
        }
        
        if(myrank==0 && kcounter==1)
        {
            cout << "printing states for mu = " << mu << endl;
            for(int i=0; i<NORB/2; ++i)
            outstates << i << '\t' << pow(abs(Hmat(2*i, NORB/2)), 2) + pow(abs(Hmat(2*i+1, NORB/2)), 2) << '\t' << pow(abs(Hmat(2*i, NORB/2-1)), 2) + pow(abs(Hmat(2*i+1, NORB/2-1)), 2) << '\t' << pow(abs(Hmat(2*i, NORB/2+1)), 2) + pow(abs(Hmat(2*i+1, NORB/2+1)), 2) << endl;
        }
        
        if(myrank==0 && kcounter==NUMMU/2)
        {
            cout << "printing states2 for mu = " << mu << endl;
            for(int i=0; i<NORB/2; ++i)
            outstates2 << i << '\t' << pow(abs(Hmat(2*i, NORB/2)), 2) + pow(abs(Hmat(2*i+1, NORB/2)), 2) << '\t' << pow(abs(Hmat(2*i, NORB/2-1)), 2) + pow(abs(Hmat(2*i+1, NORB/2-1)), 2) << '\t' << pow(abs(Hmat(2*i, NORB/2+1)), 2) + pow(abs(Hmat(2*i+1, NORB/2+1)), 2) << endl;
        }
        
        
        kcounter ++;
        
        
    }
    

    


    
    if(myrank==0)
    {
        outfile.close();
        outstates.close();
        outstates2.close();
        cout << "\n ... done! \n"<<flush;
    }
    
    
#ifndef NO_MPI
	MPI_Finalize();
#endif
    
    return 0;

}


